# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest
from unittest.mock import MagicMock, call, patch

import httpx

from pontos.git import Git
from pontos.release.main import parse_args
from pontos.release.release import ReleaseReturnValue, release
from pontos.terminal.terminal import Terminal
from pontos.testing import temp_git_repository


def mock_terminal() -> MagicMock:
    return MagicMock(spec=Terminal)


class ReleaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ["GITHUB_TOKEN"] = "foo"
        os.environ["GITHUB_USER"] = "bar"
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    @patch("pontos.release.release.Git", autospec=True)
    @patch("pontos.release.release.Path", autospec=True)
    @patch("pontos.github.api.api.httpx", autospec=True)
    @patch("pontos.github.api.release.httpx", autospec=True)
    @patch("pontos.release.release.update_version", autospec=True)
    @patch("pontos.release.release.changelog", autospec=True)
    def test_release_successfully(
        self,
        changelog_mock,
        update_version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        _git_mock,
    ):
        update_version_mock.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--git-signing-key",
                "123",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2dev",
            ]
        )

        with temp_git_repository():
            released = release(
                terminal=mock_terminal(),
                args=args,
                token=token,
            )

        self.assertEqual(released, ReleaseReturnValue.SUCCESS)

    @patch("pontos.release.release.Git", autospec=True)
    @patch("pontos.release.release.Path", autospec=True)
    @patch("pontos.github.api.api.httpx", autospec=True)
    @patch("pontos.github.api.release.httpx", autospec=True)
    @patch("pontos.release.release.update_version", autospec=True)
    @patch("pontos.release.release.changelog", autospec=True)
    def test_release_conventional_commits_successfully(
        self,
        changelog_mock,
        update_version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        _git_mock,
    ):
        update_version_mock.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--git-signing-key",
                "123",
                "--release-version",
                "1.2.3",
                "-CC",
            ]
        )
        terminal = mock_terminal()

        with temp_git_repository():
            released = release(
                terminal=terminal,
                args=args,
                token=token,
            )

        update_version_mock.assert_called_with(
            terminal,
            "1.2.4",
            develop=True,
        )
        self.assertEqual(released, ReleaseReturnValue.SUCCESS)

    @patch("pontos.release.release.Git", autospec=True)
    @patch("pontos.release.release.Path", autospec=True)
    @patch("pontos.github.api.api.httpx", autospec=True)
    @patch("pontos.github.api.release.httpx", autospec=True)
    @patch("pontos.release.release.update_version", autospec=True)
    @patch("pontos.release.release.changelog", autospec=True)
    def test_not_release_successfully_when_github_create_release_fails(
        self,
        changelog_mock,
        update_version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        _git_mock,
    ):
        update_version_mock.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 401
        fake_post.text = self.valid_gh_release_response
        fake_post.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Authorization required",
            response=fake_post,
            request=requests_mock.post,
        )
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        _, token, args = parse_args(
            [
                "release",
                "--release-version",
                "0.0.1",
            ]
        )

        with temp_git_repository():
            git = Git()
            git.add_remote("origin", "https://foo.com/bar.git")

            released = release(
                terminal=mock_terminal(),
                args=args,
                token=token,
            )
        self.assertEqual(released, ReleaseReturnValue.CREATE_RELEASE_ERROR)

    @patch("pontos.release.release.Git", autospec=True)
    @patch("pontos.release.release.Path", autospec=True)
    @patch("pontos.github.api.api.httpx", autospec=True)
    @patch("pontos.github.api.release.httpx", autospec=True)
    @patch("pontos.release.release.update_version", autospec=True)
    @patch("pontos.release.release.changelog", autospec=True)
    def test_release_to_specific_git_remote(
        self,
        changelog_mock,
        update_version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        git_mock,
    ):
        update_version_mock.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        _, token, args = parse_args(
            [
                "release",
                "--project",
                "foo",
                "--release-version",
                "0.0.1",
                "--next-version",
                "0.0.2.dev1",
                "--git-remote-name",
                "upstream",
                "--git-signing-key",
                "1234",
            ]
        )

        with temp_git_repository():
            released = release(
                terminal=mock_terminal(),
                args=args,
                token=token,
            )
        self.assertEqual(released, ReleaseReturnValue.SUCCESS)

        git_mock.return_value.push.assert_called_with(
            follow_tags=True, remote="upstream"
        )
        # git_mock.return_value.add.assert_called_with()
        git_mock.return_value.add.assert_has_calls(
            [
                call("MyProject.conf"),
                call("*__version__.py"),
                call("version.go"),
                call("CHANGELOG.md"),
            ]
        )
        git_mock.return_value.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.2.dev1\n"
            "* Add empty changelog after 0.0.1",
            verify=False,
            gpg_signing_key="1234",
        )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from pontos.models import Model


class Severity(Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class AccessVector(Enum):
    NETWORK = "NETWORK"
    ADJACENT_NETWORK = "ADJACENT_NETWORK"
    LOCAL = "LOCAL"


class AccessComplexity(Enum):
    HIGH = "HIGH"
    MEDIUM = "MEDIUM"
    LOW = "LOW"


class Authentication(Enum):
    MULTIPLE = "MULTIPLE"
    SINGLE = "SINGLE"
    NONE = "NONE"


class Impact(Enum):
    NONE = "NONE"
    PARTIAL = "PARTIAL"
    COMPLETE = "COMPLETE"


class Exploitability(Enum):
    UNPROVEN = "UNPROVEN"
    PROOF_OF_CONCEPT = "PROOF_OF_CONCEPT"
    FUNCTIONAL = "FUNCTIONAL"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class RemediationLevel(Enum):
    OFFICIAL_FIX = "OFFICIAL_FIX"
    TEMPORARY_FIX = "TEMPORARY_FIX"
    WORKAROUND = "WORKAROUND"
    UNAVAILABLE = "UNAVAILABLE"
    NOT_DEFINED = "NOT_DEFINED"


class ReportConfidence(Enum):
    UNCONFIRMED = "UNCONFIRMED"
    UNCORROBORATED = "UNCORROBORATED"
    CONFIRMED = "CONFIRMED"
    NOT_DEFINED = "NOT_DEFINED"


class CollateralDamagePotential(Enum):
    NONE = "NONE"
    LOW = "LOW"
    LOW_MEDIUM = "LOW_MEDIUM"
    MEDIUM_HIGH = "MEDIUM_HIGH"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class TargetDistribution(Enum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


class Requirement(Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    NOT_DEFINED = "NOT_DEFINED"


@dataclass
class CVSSData(Model):
    version: str
    vector_string: str
    base_score: float
    access_vector: Optional[AccessVector] = None
    access_complexity: Optional[AccessComplexity] = None
    authentication: Optional[Authentication] = None
    confidentiality_impact: Optional[Impact] = None
    integrity_impact: Optional[Impact] = None
    availability_impact: Optional[Impact] = None
    exploitability: Optional[Exploitability] = None
    remediation_level: Optional[RemediationLevel] = None
    report_confidence: Optional[ReportConfidence] = None
    temporal_score: Optional[float] = None
    collateral_damage_potential: Optional[CollateralDamagePotential] = None
    target_distribution: Optional[TargetDistribution] = None
    confidentiality_requirement: Optional[Requirement] = None
    integrity_requirement: Optional[Requirement] = None
    availability_requirement: Optional[Requirement] = None
    environmental_score: Optional[float] = None



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class UpdatedVersion:
    previous: str
    new: str


class VersionCommand(ABC):
    """Generic class usable to implement the
    version commands for several programming languages"""

    project_file_name: str

    def __init__(self, *, project_file_path: Optional[Path] = None) -> None:
        if project_file_path:
            self.project_file_path = project_file_path
        else:
            self.project_file_path = Path.cwd() / self.project_file_name

    @abstractmethod
    def get_current_version(self) -> str:
        """Get the current version of this project"""

    @abstractmethod
    def verify_version(self, version: str) -> None:
        """Verify the current version of this project"""

    @abstractmethod
    def update_version(
        self, new_version: str, *, develop: bool = False, force: bool = False
    ) -> UpdatedVersion:
        """Update the current version of this project"""

    def project_file_found(self) -> Optional[Path]:
        """
        Returns a path to the project file if a command has detected a
        corresponding project
        """
        return (
            self.project_file_path if self.project_file_path.exists() else None
        )

    def project_found(self) -> bool:
        """
        Returns True if a command has detected a corresponding project
        """
        return bool(self.project_file_found())



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import re
from pathlib import Path
from typing import Any

from pontos.version.helper import (
    VersionError,
    check_develop,
    is_version_pep440_compliant,
    safe_version,
    versions_equal,
)

from .version import UpdatedVersion, VersionCommand

GREENBONE_JS_VERSION_FILE = Path("src", "version.js")


# This class is used for JavaScript Version command(s)
class JavaScriptVersionCommand(VersionCommand):
    project_file_name = "package.json"
    _package = None

    @property
    def package(self) -> Any:
        if self._package:
            return self._package

        if not self.project_file_path.exists():
            raise VersionError(f"{self.project_file_path} file not found.")

        try:
            with self.project_file_path.open(mode="r", encoding="utf-8") as fp:
                self._package = json.load(fp)
        except OSError as e:
            raise VersionError(
                "No version tag found. Maybe this "
                "module has not been released at all."
            ) from e
        except json.JSONDecodeError as e:
            raise VersionError(
                "No valid JSON found. Maybe this "
                "module has not been released at all."
            ) from e

        if not self._package.get("version", None):
            raise VersionError(
                f"Version field missing in {self.project_file_path}."
            )

        return self._package

    def get_current_version(self) -> str:
        """Get the current version of this project
        In go the version is only defined within the repository
        tags, thus we need to check git, what tag is the latest"""
        return self.package["version"]

    def verify_version(self, version: str) -> None:
        """Verify the current version of this project"""
        current_version = self.get_current_version()
        if not is_version_pep440_compliant(current_version):
            raise VersionError(
                f"The version {current_version} is not PEP 440 compliant."
            )

        if not versions_equal(current_version, version):
            raise VersionError(
                f"Provided version {version} does not match the "
                f"current version {current_version}."
            )

    def _update_package_json(self, new_version: str) -> None:
        """
        Update the version in the package.json file
        """
        try:
            self.package["version"] = new_version

            with self.project_file_path.open(mode="w") as fp:
                json.dump(obj=self.package, fp=fp, indent=2)

        except EnvironmentError as e:
            raise VersionError(
                "No version tag found. Maybe this "
                "module has not been released at all."
            ) from e
        except json.JSONDecodeError as e:
            raise VersionError("Couldn't load JSON") from e

    def _update_version_file(self, new_version: str) -> None:
        """
        Update the version file with the new version
        """
        if GREENBONE_JS_VERSION_FILE.exists():
            content = GREENBONE_JS_VERSION_FILE.read_text(encoding="utf-8")
            content = re.sub(
                pattern=(
                    r'VERSION = (?P<quote>[\'"])[\d+\.]{2,3}'
                    r"{.dev[\d]}(?P=quote);"
                ),
                repl=f"VERSION = {new_version};",
                string=content,
            )
            GREENBONE_JS_VERSION_FILE.write_text(content, encoding="utf-8")

    def update_version(
        self, new_version: str, *, develop: bool = False, force: bool = False
    ) -> UpdatedVersion:
        new_version = safe_version(new_version)
        if not check_develop(new_version) and develop:
            new_version = f"{new_version}.dev1"

        package_version = self.get_current_version()
        if not force and versions_equal(new_version, package_version):
            return UpdatedVersion(previous=package_version, new=new_version)

        self._update_package_json(new_version=new_version)

        self._update_version_file(new_version=new_version)

        return UpdatedVersion(previous=package_version, new=new_version)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from pontos.github.models.base import Event, GitHubModel, User

__all__ = (
    "Workflow",
    "WorkflowState",
    "WorkflowRun",
    "WorkflowRunStatus",
)


@dataclass
class CommitUser(GitHubModel):
    name: str
    email: str


@dataclass
class WorkflowRunCommit(GitHubModel):
    id: str
    tree_id: str
    message: str
    timestamp: datetime
    author: Optional[CommitUser] = None
    committer: Optional[CommitUser] = None


class WorkflowState(Enum):
    ACTIVE = "active"
    DELETED = "deleted"
    DISABLED_FORK = "disabled_fork"
    DISABLED_INACTIVITY = "disabled_inactivity"
    DISABLED_MANUALLY = "disabled_manually"


@dataclass
class Workflow(GitHubModel):
    id: int
    node_id: str
    name: str
    path: str
    state: WorkflowState
    created_at: datetime
    updated_at: datetime
    url: str
    html_url: str
    badge_url: str
    deleted_at: Optional[datetime] = None


@dataclass
class WorkflowRunRepository(GitHubModel):
    id: int
    url: str
    name: str
    node_id: str
    full_name: Optional[str] = None
    owner: Optional[User] = None
    private: Optional[bool] = None
    html_url: Optional[str] = None
    description: Optional[str] = None
    fork: Optional[bool] = None
    archive_url: Optional[str] = None
    assignees_url: Optional[str] = None
    blobs_url: Optional[str] = None
    branches_url: Optional[str] = None
    collaborators_url: Optional[str] = None
    comments_url: Optional[str] = None
    commits_url: Optional[str] = None
    compare_url: Optional[str] = None
    contents_url: Optional[str] = None
    contributors_url: Optional[str] = None
    deployments_url: Optional[str] = None
    downloads_url: Optional[str] = None
    events_url: Optional[str] = None
    forks_url: Optional[str] = None
    git_commits_url: Optional[str] = None
    git_refs_url: Optional[str] = None
    git_tags_url: Optional[str] = None
    git_url: Optional[str] = None
    issue_comment_url: Optional[str] = None
    issue_events_url: Optional[str] = None
    issues_url: Optional[str] = None
    keys_url: Optional[str] = None
    labels_url: Optional[str] = None
    languages_url: Optional[str] = None
    merges_url: Optional[str] = None
    milestones_url: Optional[str] = None
    notifications_url: Optional[str] = None
    pulls_url: Optional[str] = None
    releases_url: Optional[str] = None
    ssh_url: Optional[str] = None
    stargazers_url: Optional[str] = None
    statuses_url: Optional[str] = None
    subscribers_url: Optional[str] = None
    subscription_url: Optional[str] = None
    tags_url: Optional[str] = None
    teams_url: Optional[str] = None
    trees_url: Optional[str] = None
    hooks_url: Optional[str] = None


class WorkflowRunStatus(Enum):
    ACTION_REQUIRED = "action_required"
    CANCELLED = "cancelled"
    COMPLETED = "completed"
    FAILURE = "failure"
    IN_PROGRESS = "in_progress"
    NEUTRAL = "neutral"
    QUEUED = "queued"
    REQUESTED = "requested"
    SKIPPED = "skipped"
    STALE = "stale"
    SUCCESS = "success"
    TIMED_OUT = "timed_out"
    WAITING = "waiting"
    PENDING = "pending"  # not listed in GitHub docs


@dataclass
class WorkflowRunWorkflow(GitHubModel):
    path: str
    sha: str
    ref: Optional[str] = None


@dataclass
class WorkflowRun(GitHubModel):
    artifacts_url: str
    cancel_url: str
    check_suite_url: str
    created_at: datetime
    event: Event
    head_repository: WorkflowRunRepository
    head_sha: str
    html_url: str
    id: int
    jobs_url: str
    logs_url: str
    node_id: str
    repository: WorkflowRunRepository
    rerun_url: str
    run_number: int
    updated_at: datetime
    url: str
    workflow_id: int
    workflow_url: str
    actor: Optional[User] = None
    check_suite_id: Optional[int] = None
    check_suite_node_id: Optional[str] = None
    conclusion: Optional[str] = None
    display_title: Optional[str] = None
    head_branch: Optional[str] = None
    head_commit: Optional[WorkflowRunCommit] = None
    head_repository_id: Optional[int] = None
    name: Optional[str] = None
    path: Optional[str] = None
    previous_attempt_url: Optional[str] = None
    pull_requests: List[Dict] = field(default_factory=list)
    referenced_workflows: List[WorkflowRunWorkflow] = field(
        default_factory=list
    )
    run_attempt: Optional[int] = None
    run_started_at: Optional[datetime] = None
    status: Optional[WorkflowRunStatus] = None
    triggering_actor: Optional[User] = None
